import pytest

from larch.symboltable import Group
import numpy as np

from hollowfoot import XAFSAnalysis, operation


@pytest.fixture()
def group():
    grp = Group(
        mono_energy=np.linspace(8320, 8350, num=1),
        It=np.sin(np.pi*4) + 1.5,
        I0=np.cos(np.pi*6) + 1.5,
    )
    return grp


def test_to_mu(group):
    anl = (
        XAFSAnalysis(groups=(group,))
        .to_mu("mono_energy", "It", "I0", is_transmission=True)
        .calculate()
    )
    assert anl.groups[0].mu == np.log(group.I0/group.It)

def test_to_mu_missing_keys(group):
    # What if none of the groups have the requested key
    anl = (
        XAFSAnalysis(groups=(group,))
        .to_mu("mono_energy", "Inull", "I0", is_transmission=True)
    )
    with pytest.raises(AttributeError):
        anl.calculate()



import re
from pathlib import Path
from collections.abc import Sequence, Callable

from larch.symboltable import Group
from larch.io import read_ascii


def resolve_file_paths(base: Path, glob: str = "", regex: str = "") -> list[Path]:
    """Figures out which files in a *base* path match the glob and
    regex provided.

    Also, *base* can be a path to a file, then just the base path is
    returned.

    """
    if base.is_file():
        return [base]
    # Apply glob matching
    if glob:
        children = list(base.glob(glob))
    else:
        children = list(base.iterdir())
    # Apply regex
    regex_ = re.compile(regex)
    children = [path for path in children if regex_.search(str(path))]
    return children


class NotADataFile:
    """Sentinel for if a specific file should be skipped because it has no data."""
    pass


def read_text_files(
        paths: Sequence[Path], reader: Callable[[Path, ...], Group], glob: str = "", regex="",
) -> Sequence[Group]:
    """Iterate data groups from text files.

    Useful for making beamline-specific input functions.

    Parameters
    ==========
    base
      Path objects that will be opened and read for data.
    reader
      The function that knows how to load data in this specific format.

    """
    for path in paths:
        maybe_group = reader(path)
        if isinstance(maybe_group, NotADataFile):
            continue
        yield maybe_group


def read_aps_20bmb(base: str | Path, glob: str = "", regex: str = "") -> list[Group]:
    """Read XAFS data measured at APS beamline 20-BM-B.

    The first argument can be either a specific file to read, or a
    directory containing such files.

    Selecting specific files from a directory can be accomplished
    using either globs or regular expressions:

    .. code-block:: python
    
        read_aps_20bmb_

    Parameters
    ==========
    base
      A filesystem path in which to look for files, or else a
      specific file to read.
    reader
      The function that knows how to load data in this specific format.
    glob
      If *base* is a directory, this glob will be used as a pattern
      for restricting files.
    regex
      If *base* is a directory, only files matching this regular
      expression will be read.

    """
    paths = resolve_file_paths(Path(base), glob=glob, regex=regex)
    
    def reader(fp):
        if fp.suffix == ".last":
            return NotADataFile()
        return read_ascii(fp)

    groups = list(read_text_files(paths, reader))
    return groups



from pathlib import Path

from hollowfoot.readers import read_aps_20bmb, resolve_file_paths

data_dir = Path(__file__).parent / "data"


def test_resolve_single_file(tmp_path):
    # Set up some files to check
    good_path = tmp_path / "spam.eggs"
    good_path.touch()
    resolved = resolve_file_paths(good_path)
    assert resolved == [good_path]


def test_resolve_directory(tmp_path):
    # Set up some files to check
    good_path = tmp_path / "spam.eggs"
    good_path.touch()
    bad_path = tmp_path / "spam.parrot"
    bad_path.touch()
    resolved = resolve_file_paths(tmp_path)
    assert set(resolved) == {good_path, bad_path}


def test_resolve_file_globs(tmp_path):
    # Set up some files to check
    good_path = tmp_path / "spam.eggs"
    good_path.touch()
    bad_path = tmp_path / "spam.parrot"
    bad_path.touch()
    resolved = resolve_file_paths(tmp_path, glob="s*s")
    assert resolved == [good_path]


def test_resolve_file_regex(tmp_path):
    # Set up some files to check
    good_path = tmp_path / "spam.eggs"
    good_path.touch()
    bad_path = tmp_path / "spamparrot"
    bad_path.touch()
    resolved = resolve_file_paths(tmp_path, regex=r"spam\..+")
    assert resolved == [good_path]


def test_read_aps_20bmb(tmp_path):
    data_file = data_dir / "Ni-foil-EXAFS.0002"
    groups = read_aps_20bmb(data_file)
    assert len(groups) == 1



from pathlib import Path
from dataclasses import dataclass
from collections.abc import Sequence, Iterable, Callable, Mapping
from functools import wraps
from typing import Any
import warnings

from larch.symboltable import Group

from .readers import read_aps_20bmb

@dataclass(frozen=True)
class Operation:
    """A discrete computational unit of analysis."""

    desc: str
    func: Callable[[Sequence[Group], ...], Sequence[Group]]
    args: tuple[Any, ...]
    kwargs: Mapping[Any, Any]


def operation(desc: str):
    """A decorator that"""

    def wrapper(fn: Callable[[Sequence[Group], ...], Sequence[Group]]):
        @wraps(fn)
        def inner(analysis: "BaseAnalysis", *args, **kwargs) -> "BaseAnalysis":
            new_operation = Operation(desc=desc, func=fn, args=args, kwargs=kwargs)
            new_analysis = type(analysis)(
                groups=analysis.groups,
                operations=(*analysis.operations, new_operation),
                past_operations=analysis.past_operations,
            )
            return new_analysis

        return inner

    return wrapper


class Analysis:
    groups: Sequence[Group]
    operations: tuple[Operation, ...]

    def __init__(
        self,
        groups: Iterable[Group] = (),
        operations: tuple[Operation, ...] = (),
        past_operations: tuple[Operation, ...] = (),
    ):
        self.groups = groups
        self.operations = operations
        self.past_operations = past_operations

    def calculate(self):
        """Apply all pending operations and produce a new analysis object."""
        groups = self.groups
        operations = self.operations
        for op in self.operations:
            groups = op.func(list(groups), *op.args, **op.kwargs)
        groups = tuple(groups)
        if len(groups) == 0:
            warnings.warn(f"Operation {op} produced 0 valid groups")
        return type(self)(
            tuple(groups),
            operations=[],
            past_operations=(*self.past_operations, *operations),
        )

    @classmethod
    def from_aps_20bmb(cls, base: str | Path, glob: str = "", regex: str = "") -> "Analysis":
        """Read XAFS data measured at APS beamline 20-BM-B.

        The first argument can be either a specific file to read, or a
        directory containing such files.

        Selecting specific files from a directory can be accomplished
        using either globs or regular expressions:

        .. code-block:: python

            read_aps_20bmb_

        Parameters
        ==========
        base
          A filesystem path in which to look for files, or else a
          specific file to read.
        reader
          The function that knows how to load data in this specific format.
        glob
          If *base* is a directory, this glob will be used as a pattern
          for restricting files.
        regex
          If *base* is a directory, only files matching this regular
          expression will be read.

        """
        groups = read_aps_20bmb(base=base, glob=glob, regex=regex)
        return cls(groups=groups)



import warnings
from collections.abc import Callable, Iterable, Sequence
from copy import copy
from functools import wraps
from pathlib import Path
from collections.abc import Callable, Mapping, Sequence
from dataclasses import dataclass
from functools import wraps
from typing import Any

import numpy as np
from larch.io import merge_groups, read_ascii
from larch.plot import bokeh_xafsplots as xafsplots
from larch.symboltable import Group
from larch.xafs import pre_edge, autobk
from larch.symboltable import Group

from hollowfoot.analysis import Analysis, operation


class XAFSAnalysis(Analysis):
    def is_flattened(self):
        print([op.func for op in self.past_operations])
        return self.fit_edge_jump.__wrapped__ in [
            op.func for op in self.past_operations
        ]

    @operation(desc="Calculate µ(E)")
    def to_mu(
        groups: Sequence[Group],
        energy: str,
        signal: str,
        reference: str | None = None,
        is_transmission=False,
    ):
        
        for group in groups:
            ydata = getattr(group, signal)
            rdata = getattr(group, reference) if reference is not None else 1
            mu = ydata / rdata
            if is_transmission:
                mu = -np.log(mu)
            yield Group(mu=mu, energy=getattr(group, energy))

    def summarize(self):
        new_analysis = self.calculate()
        print("Steps:")
        for operation in new_analysis.past_operations:
            print(f"- {operation.desc}")
        return new_analysis

    @operation(desc="merge data groups")
    def merge(groups, *args, **kwargs):
        return [merge_groups(groups, *args, **kwargs)]

    @wraps(xafsplots.plot_mu)
    def plot_mu(self, **kwargs):
        new_analysis = self.calculate()
        if "fig" in kwargs:
            fig = kwargs["fig"]
        else:
            fig = kwargs["fig"] = xafsplots.BokehFigure()
        # Get some default arguments
        kwargs.setdefault("show_flat", new_analysis.is_flattened())
        # Do the actual plotting
        for group in new_analysis.groups:
            fig = xafsplots.plot_mu(group, show=False, **kwargs)
        fig.show(show=True)
        return new_analysis

    @wraps(xafsplots.plot_chik)
    def plot_chik(self, **kwargs):
        new_analysis = self.calculate()
        if "fig" in kwargs:
            fig = kwargs["fig"]
        else:
            fig = kwargs["fig"] = xafsplots.BokehFigure()
        # Do the actual plotting
        for group in new_analysis.groups:
            fig = xafsplots.plot_chik(group, show=False, **kwargs)
        fig.show(show=True)
        return new_analysis

    @operation(desc="fit edge jump in µ(E)")
    def fit_edge_jump(groups, *args, **kwargs):
        for group in groups:
            new_group = copy(group)
            pre_edge(new_group, *args, **kwargs)
            yield new_group

    @operation(desc="subtract background to produce χ(k)")
    def subtract_background(groups, *args, **kwargs):
        for group in groups:
            new_group = copy(group)
            autobk(new_group, *args, **kwargs)
            yield new_group




from pathlib import Path

from larch.symboltable import Group
import numpy as np

from hollowfoot import Analysis, operation

data_dir = Path(__file__).parent / "data"


class TestAnalysis(Analysis):
    @operation(desc="run test code")
    def noop(self, color="red"):
        pass


def test_analysis_chaining():
    group = Group(x=np.linspace(0, 100, num=19))
    anl = TestAnalysis((group,))
    new_anl = anl.noop()
    assert isinstance(new_anl, Analysis)
    assert new_anl is not anl


def test_operation_adds_to_stack():
    anl = TestAnalysis().noop()
    assert len(anl.operations) == 1
    assert anl.operations[0].desc == "run test code"


def test_read_aps_20bmb(tmp_path):
    data_file = data_dir / "Ni-foil-EXAFS.0002"
    analysis = Analysis.from_aps_20bmb(data_file)
    assert len(analysis.groups) == 1



from ._version import get_versions
from .analysis import Analysis, operation  # noqa: F401
from .xafs_analysis import XAFSAnalysis  # noqa: F401


__version__ = get_versions()["version"]
del get_versions

# TODO: fill this in with appropriate star imports:
__all__ = ["Analysis", "operation"]
